﻿Option Explicit On
Option Strict On

Imports System.ComponentModel
Imports System.Windows.Forms.VisualStyles

Imports MvpsOrg.Dotnet.Win32.Transactions.IO

Public Class CopyFileControl
    Private m_BorderRenderer As New VisualStyleRenderer(VisualStyleElement.Button.GroupBox.Normal)
    Private m_Cancel As Boolean

    Private Delegate Sub UpdateProgressBarInvoker(ByVal Value As Integer)
    Private Delegate Sub UpdateStatusInvoker(ByVal Text As String)

    Private Sub CopyFileControl_Paint(ByVal sender As Object, ByVal e As PaintEventArgs) Handles Me.Paint
        m_BorderRenderer.DrawBackground(e.Graphics, Me.ClientRectangle)
    End Sub

    Private Sub ButtonStart_Click(ByVal sender As Object, ByVal e As EventArgs) Handles ButtonStart.Click
        Me.Label3.Text = "Copying file in progress…"
        Me.TextBoxDestFileName.ReadOnly = True
        Me.TextBoxSourceFileName.ReadOnly = True
        Me.ButtonStart.Enabled = False
        Me.ButtonBrowseSourceFileName.Enabled = False
        Me.ButtonBrowseDestFileName.Enabled = False
        Me.ButtonCancel.Enabled = True
        Me.BackgroundWorker.RunWorkerAsync()
    End Sub

    Private Sub ButtonCancel_Click(ByVal sender As Object, ByVal e As EventArgs) Handles ButtonCancel.Click
        Me.ButtonCancel.Enabled = False
        m_Cancel = True
    End Sub

    Private Sub ButtonRemove_Click(ByVal sender As Object, ByVal e As EventArgs) Handles ButtonRemove.Click
        Me.Parent.Controls.Remove(Me)
    End Sub

    Private Sub ButtonBrowseSourceFileName_Click(ByVal sender As Object, ByVal e As EventArgs) Handles ButtonBrowseSourceFileName.Click
        If Me.BrowseFileDialog.ShowDialog = DialogResult.OK Then
            Me.TextBoxSourceFileName.Text = Me.BrowseFileDialog.FileName
        End If
    End Sub

    Private Sub ButtonBrowseDestFileName_Click(ByVal sender As Object, ByVal e As EventArgs) Handles ButtonBrowseDestFileName.Click
        If Me.BrowseFileDialog.ShowDialog = DialogResult.OK Then
            Me.TextBoxDestFileName.Text = Me.BrowseFileDialog.FileName
        End If
    End Sub

    Private Sub BackgroundWorker_DoWork(ByVal sender As Object, ByVal e As DoWorkEventArgs) Handles BackgroundWorker.DoWork
        Try
            File.Copy( _
                Me.TextBoxSourceFileName.Text, _
                Me.TextBoxDestFileName.Text, _
                AddressOf CopyProgressCallback, _
                CopyFileOptions.None _
            )
        Catch ex As Exception
            Me.UpdateStatus(MakeSentence(ex.Message))
            Return
        End Try
        Me.UpdateStatus("File copied successfully.")
    End Sub

    Private Sub BackgroundWorker_RunWorkerCompleted(ByVal sender As Object, ByVal e As RunWorkerCompletedEventArgs) Handles BackgroundWorker.RunWorkerCompleted
        Me.ButtonRemove.Visible = True
        Me.ButtonCancel.Enabled = False
    End Sub

    Private Sub UpdateProgressBar(ByVal Value As Integer)
        If Me.ProgressBar.InvokeRequired Then
            Me.ProgressBar.BeginInvoke( _
                New UpdateProgressBarInvoker(AddressOf UpdateProgressBar), _
                New Object() {Value} _
            )
        Else
            Me.ProgressBar.Value = Value
        End If
    End Sub

    Private Sub UpdateStatus(ByVal Text As String)
        If Me.Label3.InvokeRequired Then
            Me.Label3.BeginInvoke( _
                New UpdateStatusInvoker(AddressOf UpdateStatus), _
                New Object() {Text} _
            )
        Else
            Me.Label3.Text = Text
        End If
    End Sub

    Private Function CopyProgressCallback( _
        ByVal TotalFileSize As Long, _
        ByVal TotalBytesTransferred As Long, _
        ByVal StreamSize As Long, _
        ByVal StreamBytesTransferred As Long, _
        ByVal StreamNumber As Integer, _
        ByVal CallbackReason As CallbackReason _
    ) As CopyProgressResult
        If m_Cancel Then
            Me.UpdateProgressBar(0)
            Return CopyProgressResult.Cancel
        End If
        Me.UpdateProgressBar(CInt((TotalBytesTransferred / TotalFileSize) * 100))
        Return CopyProgressResult.Continue
    End Function

    Private Shared Function MakeSentence(ByVal Text As String) As String
        If Not Text.EndsWith(".") Then
            Return Text & "."
        Else
            Return Text
        End If
    End Function
End Class