﻿Option Explicit On
Option Strict On

Imports System.Diagnostics.CodeAnalysis
Imports System.Runtime.InteropServices

Imports Microsoft.Win32.SafeHandles
Imports MvpsOrg.Dotnet.Win32.SafeHandles

Namespace MvpsOrg.Dotnet.Win32
    Friend NotInheritable Class NativeMethods
        Private Sub New()
            '
        End Sub

#Region "Kernel Transaction Manager"
        Friend Declare Unicode Function CreateTransaction Lib "KtmW32.dll" ( _
            ByVal lpTransactionAttributes As IntPtr, _
            ByVal UOW As IntPtr, _
            ByVal CreateOptions As Int32, _
            ByVal IsolationLevel As Int32, _
            ByVal IsolationFlags As Int32, _
            ByVal Timeout As Int32, _
            ByVal Description As String _
        ) As SafeTransactionHandle

        Friend Declare Function OpenTransaction Lib "KtmW32.dll" ( _
            ByVal dwDesiredAccess As TRANSACTION_ACCESS_Flags, _
            ByRef TransactionId As GUID _
        ) As SafeTransactionHandle

        Friend Declare Function CommitTransaction Lib "KtmW32.dll" ( _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Boolean

        Friend Declare Function CommitTransactionAsync Lib "KtmW32.dll" ( _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Boolean

        Friend Declare Function RollbackTransaction Lib "KtmW32.dll" ( _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Boolean

        Friend Declare Function RollbackTransactionAsync Lib "KtmW32.dll" ( _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Boolean

        Friend Declare Function RollbackComplete Lib "KtmW32.dll" ( _
            ByVal EnlistmentHandle As IntPtr, _
            ByRef TmVirtualClock As Int64 _
        ) As Boolean

        Friend Declare Function GetTransactionId Lib "KtmW32.dll" ( _
            ByVal TransactionHandle As SafeTransactionHandle, _
            ByRef TransactionId As GUID _
        ) As Boolean

        Friend Declare Unicode Function GetTransactionInformation Lib "KtmW32.dll" ( _
            ByVal TransactionHandle As SafeTransactionHandle, _
            ByRef Outcome As Int32, _
            ByRef IsolationLevel As Int32, _
            ByRef IsolationFlags As Int32, _
            ByRef Timeout As Int32, _
            ByVal BufferLength As Int32, _
            ByVal Description As String _
        ) As Boolean

        Friend Declare Unicode Function SetTransactionInformation Lib "KtmW32.dll" ( _
            ByVal TransactionHandle As SafeTransactionHandle, _
            ByVal IsolationLevel As Int32, _
            ByVal IsolationFlags As Int32, _
            ByVal Timeout As Int32, _
            ByVal Description As String _
        ) As Boolean

        <StructLayout(LayoutKind.Sequential)> _
        Friend Structure GUID
            Friend Data1 As Int32
            Friend Data2 As Int16
            Friend Data3 As Int16
            <MarshalAs(UnmanagedType.ByValArray, SizeConst:=8)> _
            Friend Data4() As Byte
        End Structure

        <Flags()> _
        Friend Enum TRANSACTION_ACCESS_Flags
            TRANSACTION_QUERY_INFORMATION = &H1
            TRANSACTION_SET_INFORMATION = &H2
            TRANSACTION_ENLIST = &H4
            TRANSACTION_COMMIT = &H8
            TRANSACTION_ROLLBACK = &H10
            TRANSACTION_PROPAGATE = &H20
            TRANSACTION_GENERIC_READ = &H120001
            TRANSACTION_GENERIC_WRITE = &H12003E
            TRANSACTION_GENERIC_EXECUTE = &H120018
            TRANSACTION_ALL_ACCESS = &H12003F
            TRANSACTION_RESOURCE_MANAGER_RIGHTS = &H120037
        End Enum
#End Region

#Region "CreateFile"
        Friend Declare Auto Function CreateFile Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal dwDesiredAccess As DesiredAccess_Flags, _
            ByVal dwShareMode As FILE_SHARE_Flags, _
            ByRef lpSecurityAttributes As SECURITY_ATTRIBUTES, _
            ByVal dwCreationDisposition As CreationDisposition_Constants, _
            ByVal dwFlagsAndAttributes As FlagsAndAttributes_Flags, _
            ByVal hTemplateFile As SafeFileHandle _
        ) As SafeFileHandle

        Friend Declare Auto Function CreateFileTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal dwDesiredAccess As DesiredAccess_Flags, _
            ByVal dwShareMode As FILE_SHARE_Flags, _
            ByRef lpSecurityAttributes As SECURITY_ATTRIBUTES, _
            ByVal dwCreationDisposition As CreationDisposition_Constants, _
            ByVal dwFlagsAndAttributes As FlagsAndAttributes_Flags, _
            ByVal hTemplateFile As SafeFileHandle, _
            ByVal hTransaction As SafeTransactionHandle, _
            ByRef pusMiniVersion As TXFS_MINIVERSION_Constants, _
            ByVal pExtendedParameter As IntPtr _
        ) As SafeFileHandle

        Friend Declare Auto Function CreateFileTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal dwDesiredAccess As DesiredAccess_Flags, _
            ByVal dwShareMode As FILE_SHARE_Flags, _
            ByRef lpSecurityAttributes As SECURITY_ATTRIBUTES, _
            ByVal dwCreationDisposition As CreationDisposition_Constants, _
            ByVal dwFlagsAndAttributes As FlagsAndAttributes_Flags, _
            ByVal hTemplateFile As SafeFileHandle, _
            ByVal hTransaction As SafeTransactionHandle, _
            ByVal pusMiniVersion As IntPtr, _
            ByVal pExtendedParameter As IntPtr _
        ) As SafeFileHandle

        <Flags()> _
        Friend Enum DesiredAccess_Flags
            QueryDeviceOnly = &H0
            GENERIC_READ = &H80000000
            GENERIC_WRITE = &H40000000
            GENERIC_EXECUTE = &H200000
            GENERIC_ALL = &H10000000
            FILE_READ_DATA = &H1            ' file & pipe
            FILE_LIST_DIRECTORY = &H1       ' directory
            FILE_WRITE_DATA = &H2           ' file & pipe
            FILE_ADD_FILE = &H2             ' directory
            FILE_APPEND_DATA = &H4          ' file
            FILE_ADD_SUBDIRECTORY = &H4     ' directory
            FILE_CREATE_PIPE_INSTANCE = &H4 ' named pipe
            FILE_READ_EA = &H8              ' file & directory
            FILE_WRITE_EA = &H10            ' file & directory
            FILE_EXECUTE = &H20             ' file
            FILE_TRAVERSE = &H20            ' directory
            FILE_DELETE_CHILD = &H40        ' directory
            FILE_READ_ATTRIBUTES = &H80     ' all
            FILE_WRITE_ATTRIBUTES = &H100   ' all
            FILE_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED Or SYNCHRONIZE Or &H1FF
            FILE_GENERIC_READ = STANDARD_RIGHTS_READ Or FILE_READ_DATA Or FILE_READ_ATTRIBUTES Or FILE_READ_EA Or SYNCHRONIZE
            FILE_GENERIC_WRITE = STANDARD_RIGHTS_WRITE Or FILE_WRITE_DATA Or FILE_WRITE_ATTRIBUTES Or FILE_WRITE_EA Or FILE_APPEND_DATA Or SYNCHRONIZE
            FILE_GENERIC_EXECUTE = STANDARD_RIGHTS_EXECUTE Or FILE_READ_ATTRIBUTES Or FILE_EXECUTE Or SYNCHRONIZE
            DELETE = &H10000
            READ_CONTROL = &H20000
            WRITE_DAC = &H40000
            WRITE_OWNER = &H80000
            SYNCHRONIZE = &H100000
            STANDARD_RIGHTS_REQUIRED = &HF0000
            STANDARD_RIGHTS_READ = READ_CONTROL
            STANDARD_RIGHTS_WRITE = READ_CONTROL
            STANDARD_RIGHTS_EXECUTE = READ_CONTROL
            STANDARD_RIGHTS_ALL = &H1F0000
            SPECIFIC_RIGHTS_ALL = &HFFFF
        End Enum

        <Flags()> _
        Friend Enum FILE_SHARE_Flags
            NotShared = &H0
            FILE_SHARE_READ = &H1
            FILE_SHARE_WRITE = &H2
            FILE_SHARE_DELETE = &H4
        End Enum

        Friend Enum CreationDisposition_Constants
            OPEN_EXISTING = 1
            CREATE_NEW = 2
            OPEN_ALWAYS = 3
            TRUNCATE_EXISTING = 4
            CREATE_ALWAYS = 5
        End Enum

        <Flags()> _
        Friend Enum FlagsAndAttributes_Flags
            None = 0
            FILE_ATTRIBUTE_READONLY = &H1
            FILE_ATTRIBUTE_HIDDEN = &H2
            FILE_ATTRIBUTE_SYSTEM = &H4
            FILE_ATTRIBUTE_DIRECTORY = &H10
            FILE_ATTRIBUTE_ARCHIVE = &H20
            FILE_ATTRIBUTE_DEVICE = &H40
            FILE_ATTRIBUTE_NORMAL = &H80
            FILE_ATTRIBUTE_TEMPORARY = &H100
            FILE_ATTRIBUTE_SPARSE_FILE = &H200
            FILE_ATTRIBUTE_REPARSE_POINT = &H400
            FILE_ATTRIBUTE_COMPRESSED = &H800
            FILE_ATTRIBUTE_OFFLINE = &H1000
            FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = &H2000
            FILE_ATTRIBUTE_ENCRYPTED = &H4000
            FILE_ATTRIBUTE_VIRTUAL = &H10000
            FILE_FLAG_WRITE_THROUGH = &H80000000
            FILE_FLAG_OVERLAPPED = &H40000000
            FILE_FLAG_NO_BUFFERING = &H20000000
            FILE_FLAG_RANDOM_ACCESS = &H10000000
            FILE_FLAG_SEQUENTIAL_SCAN = &H8000000
            FILE_FLAG_DELETE_ON_CLOSE = &H4000000
            FILE_FLAG_BACKUP_SEMANTICS = &H2000000
            FILE_FLAG_POSIX_SEMANTICS = &H1000000
            FILE_FLAG_OPEN_REPARSE_POINT = &H200000
            FILE_FLAG_OPEN_NO_RECALL = &H100000
            SECURITY_ANONYMOUS = &H0
            SECURITY_IDENTIFICATION = &H10000
            SECURITY_IMPERSONATION = &H20000
            SECURITY_DELEGATION = &H30000
            SECURITY_CONTEXT_TRACKING = &H40000
            SECURITY_EFFECTIVE_ONLY = &H80000
        End Enum

        Friend Enum TXFS_MINIVERSION_Constants As UInt16
            TXFS_MINIVERSION_COMMITTED_VIEW = &H0
            TXFS_MINIVERSION_DIRTY_VIEW = &HFFFE
            TXFS_MINIVERSION_DEFAULT_VIEW = &HFFFF
        End Enum
#End Region

#Region "DeleteFile"
        Friend Declare Auto Function DeleteFile Lib "Kernel32.dll" ( _
            ByVal lpFileName As String _
        ) As Boolean

        Friend Declare Auto Function DeleteFileTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean
#End Region

#Region "CopyFile"
        Friend Declare Auto Function CopyFileEx Lib "Kernel32.dll" ( _
            ByVal lpExistingFileName As String, _
            ByVal lpNewFileName As String, _
            ByVal lpProgressRoutine As CopyProgressRoutine, _
            ByVal lpData As IntPtr, _
            ByRef pbCancel As Boolean, _
            ByVal dwCopyFlags As COPY_FILE_Flags _
        ) As Boolean

        Friend Declare Auto Function CopyFileEx Lib "Kernel32.dll" ( _
            ByVal lpExistingFileName As String, _
            ByVal lpNewFileName As String, _
            ByVal lpProgressRoutine As CopyProgressRoutine, _
            ByVal lpData As IntPtr, _
            ByVal pbCancel As IntPtr, _
            ByVal dwCopyFlags As COPY_FILE_Flags _
        ) As Boolean

        Friend Declare Auto Function CopyFileTransacted Lib "Kernel32.dll" ( _
            ByVal lpExistingFileName As String, _
            ByVal lpNewFileName As String, _
            ByVal lpProgressRoutine As CopyProgressRoutine, _
            ByVal lpData As IntPtr, _
            ByRef pbCancel As Boolean, _
            ByVal dwCopyFlags As COPY_FILE_Flags, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean

        Friend Declare Auto Function CopyFileTransacted Lib "Kernel32.dll" ( _
            ByVal lpExistingFileName As String, _
            ByVal lpNewFileName As String, _
            ByVal lpProgressRoutine As CopyProgressRoutine, _
            ByVal lpData As IntPtr, _
            ByVal pbCancel As IntPtr, _
            ByVal dwCopyFlags As COPY_FILE_Flags, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean

        Friend Delegate Function CopyProgressRoutine( _
            ByVal TotalFileSize As Int64, _
            ByVal TotalBytesTransferred As Int64, _
            ByVal StreamSize As Int64, _
            ByVal StreamBytesTransferred As Int64, _
            ByVal dwStreamNumber As Int32, _
            ByVal dwCallbackReason As CALLBACK_Constants, _
            ByVal hSourceFile As IntPtr, _
            ByVal hDestinationFile As IntPtr, _
            ByVal lpData As IntPtr _
        ) As CopyProgressRoutine_Result

        Friend Enum CALLBACK_Constants
            CALLBACK_CHUNK_FINISHED = &H0
            CALLBACK_STREAM_SWITCH = &H1
        End Enum

        Friend Enum CopyProgressRoutine_Result
            PROGRESS_CONTINUE = 0
            PROGRESS_CANCEL = 1
            PROGRESS_STOP = 2
            PROGRESS_QUIET = 3
        End Enum

        <Flags()> _
        Friend Enum COPY_FILE_Flags
            None = &H0
            COPY_FILE_FAIL_IF_EXISTS = &H1
            COPY_FILE_RESTARTABLE = &H2
            COPY_FILE_OPEN_SOURCE_FOR_WRITE = &H4
            COPY_FILE_ALLOW_DECRYPTED_DESTINATION = &H8
        End Enum
#End Region

#Region "MoveFile"
        Friend Declare Auto Function MoveFileEx Lib "Kernel32.dll" ( _
            ByVal lpExistingFileName As String, _
            ByVal lpNewFileName As String, _
            ByVal dwFlags As MOVEFILE_Flags _
        ) As Boolean

        Friend Declare Auto Function MoveFileWithProgress Lib "Kernel32.dll" ( _
            ByVal lpExistingFileName As String, _
            ByVal lpNewFileName As String, _
            ByVal lpProgressRoutine As CopyProgressRoutine, _
            ByVal lpData As IntPtr, _
            ByVal dwFlags As MOVEFILE_Flags _
        ) As Boolean

        Friend Declare Auto Function MoveFileTransacted Lib "Kernel32.dll" ( _
            ByVal lpExistingFileName As String, _
            ByVal lpNewFileName As String, _
            ByVal lpProgressRoutine As CopyProgressRoutine, _
            ByVal lpData As IntPtr, _
            ByVal dwFlags As Int32, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean

        <Flags()> _
        Friend Enum MOVEFILE_Flags
            None = &H0
            MOVEFILE_COPY_ALLOWED = &H2
            MOVEFILE_CREATE_HARDLINK = &H10
            MOVEFILE_DELAY_UNTIL_REBOOT = &H4
            MOVEFILE_FAIL_IF_NOT_TRACKABLE = &H20
            MOVEFILE_REPLACE_EXISTING = &H1
            MOVEFILE_WRITE_THROUGH = &H8
        End Enum
#End Region

#Region "CreateHardLink"
        Friend Declare Auto Function CreateHardLink Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal lpExistingFileName As String, _
            ByVal lpSecurityAttributes As SECURITY_ATTRIBUTES _
        ) As Boolean

        Friend Declare Auto Function CreateHardLinkTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal lpExistingFileName As String, _
            ByVal lpSecurityAttributes As SECURITY_ATTRIBUTES, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean
#End Region

#Region "CreateSymbolicLink"
        Friend Declare Auto Function CreateSymbolicLink Lib "Kernel32.dll" ( _
            ByVal lpSymlinkFileName As String, _
            ByVal lpTargetFileName As String, _
            ByVal dwFlags As SYMBOLIC_LINK_FLAG_Flags _
        ) As Boolean

        Friend Declare Auto Function CreateSymbolicLinkTransacted Lib "Kernel32.dll" ( _
            ByVal lpSymlinkFileName As String, _
            ByVal lpTargetFileName As String, _
            ByVal dwFlags As SYMBOLIC_LINK_FLAG_Flags, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean

        Friend Enum SYMBOLIC_LINK_FLAG_Flags
            SYMBOLIC_LINK_FLAG_FILE = &H0   ' User-defined name.
            SYMBOLIC_LINK_FLAG_DIRECTORY = &H1
        End Enum
#End Region

#Region "GetFileAttributes"
        Friend Declare Function GetFileAttributes Lib "Kernel32.dll" ( _
            ByVal lpFileName As String _
        ) As FILE_ATTRIBUTE_Flags

        Friend Declare Auto Function GetFileAttributesTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal fInfoLevelId As GET_FILEEX_INFO_LEVELS, _
            ByRef lpFileInformation As WIN32_FILE_ATTRIBUTE_DATA, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean

        <StructLayout(LayoutKind.Sequential)> _
        Friend Structure WIN32_FILE_ATTRIBUTE_DATA
            Friend dwFileAttributes As FILE_ATTRIBUTE_Flags
            Friend ftCreationTime As ComTypes.FILETIME
            Friend ftLastAccessTime As ComTypes.FILETIME
            Friend ftLastWriteTime As ComTypes.FILETIME
            Friend nFileSizeHigh As Int32
            Friend nFileSizeLow As Int32
        End Structure

        Friend Enum GET_FILEEX_INFO_LEVELS
            GetFileExInfoStandard
            GetFileExMaxInfoLevelOne
        End Enum

        <Flags()> _
        Friend Enum FILE_ATTRIBUTE_Flags
            FILE_ATTRIBUTE_ARCHIVE = &H20
            FILE_ATTRIBUTE_COMPRESSED = &H800
            FILE_ATTRIBUTE_DEVICE = &H40
            FILE_ATTRIBUTE_DIRECTORY = &H10
            FILE_ATTRIBUTE_ENCRYPTED = &H4000
            FILE_ATTRIBUTE_HIDDEN = &H2
            FILE_ATTRIBUTE_NORMAL = &H80
            FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = &H2000
            FILE_ATTRIBUTE_OFFLINE = &H1000
            FILE_ATTRIBUTE_READONLY = &H1
            FILE_ATTRIBUTE_REPARSE_POINT = &H400
            FILE_ATTRIBUTE_SPARSE_FILE = &H200
            FILE_ATTRIBUTE_SYSTEM = &H4
            FILE_ATTRIBUTE_TEMPORARY = &H100
            FILE_ATTRIBUTE_VIRTUAL = &H10000
        End Enum

        Friend Const INVALID_FILE_ATTRIBUTES As Int32 = -1
#End Region

#Region "SetFileAttributes"
        Friend Declare Ansi Function SetFileAttributes Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal dwFileAttributes As FILE_ATTRIBUTE_Flags _
        ) As Boolean

        Friend Declare Auto Function SetFileAttributesTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal dwFileAttributes As Int32, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean
#End Region

#Region "GetCompressedFileSize"
        Friend Declare Auto Function GetCompressedFileSize Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByRef lpFileSizeHigh As Int32 _
        ) As Int32

        Friend Declare Auto Function GetCompressedFileSizeTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByRef lpFileSizeHigh As Int32, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Int32

        Friend Const INVALID_FILE_SIZE As Int32 = -1
#End Region

#Region "CreateDirectory"
        Friend Declare Auto Function CreateDirectoryEx Lib "Kernel32.dll" ( _
            ByVal lpTemplateDirectory As String, _
            ByVal lpNewDirectory As String, _
            ByRef lpSecurityAttributes As SECURITY_ATTRIBUTES _
        ) As Boolean

        Friend Declare Auto Function CreateDirectoryTransacted Lib "Kernel32.dll" ( _
            ByVal lpTemplateDirectory As String, _
            ByVal lpNewDirectory As String, _
            ByRef lpSecurityAttributes As SECURITY_ATTRIBUTES, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean
#End Region

#Region "RemoveDirectory"
        Friend Declare Auto Function RemoveDirectory Lib "Kernel32.dll" ( _
            ByVal lpPathName As String _
        ) As Boolean

        Friend Declare Auto Function RemoveDirectoryTransacted Lib "Kernel32.dll" ( _
            ByVal lpPathName As String, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Boolean
#End Region

#Region "Find*File"
        Friend Declare Auto Function FindFirstFileEx Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal fInfoLevelId As FINDEX_INFO_LEVELS, _
            ByRef lpFindFileData As WIN32_FIND_DATA, _
            ByVal fSearchOp As FINDEX_SEARCH_OPS, _
            ByVal lpSearchFilter As IntPtr, _
            ByVal dwAdditionalFlags As AdditionalFlags_Flags _
        ) As SafeFindHandle

        Friend Declare Auto Function FindFirstFileTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal fInfoLevelId As FINDEX_INFO_LEVELS, _
            ByRef lpFindFileData As WIN32_FIND_DATA, _
            ByVal fSearchOp As FINDEX_SEARCH_OPS, _
            ByVal lpSearchFilter As IntPtr, _
            ByVal dwAdditionalFlags As AdditionalFlags_Flags, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As SafeFindHandle

        Friend Declare Auto Function FindNextFile Lib "Kernel32.dll" ( _
            ByVal hFindFile As SafeFindHandle, _
            ByRef lpFindFileData As WIN32_FIND_DATA _
        ) As Boolean

        Friend Declare Function FindClose Lib "Kernel32.dll" ( _
             ByRef hFindFile As IntPtr _
        ) As Boolean

        <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
        Friend Structure WIN32_FIND_DATA
            Friend dwFileAttributes As FILE_ATTRIBUTE_Flags
            Friend ftCreationTime As ComTypes.FILETIME
            Friend ftLastAccessTime As ComTypes.FILETIME
            Friend ftLastWriteTime As ComTypes.FILETIME
            Friend nFileSizeHigh As Int32
            Friend nFileSizeLow As Int32
            Friend dwReserved0 As Int32
            Friend dwReserved1 As Int32
            <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=MAX_PATH)> _
            Friend cFileName As String
            <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=14)> _
            Friend cAlternateFileName As String
        End Structure

        Friend Enum FINDEX_INFO_LEVELS
            FindExInfoStandard
            FindExInfoMaxInfoLevel
        End Enum

        Friend Enum FINDEX_SEARCH_OPS
            FindExSearchNameMatch
            FindExSearchLimitToDirectories
            FindExSearchLimitToDevices
        End Enum

        Friend Enum AdditionalFlags_Flags
            None = 0
            FIND_FIRST_EX_CASE_SENSITIVE = 1
        End Enum
#End Region

#Region "Find*FileNameW"
        Friend Declare Unicode Function FindFirstFileNameW Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal dwFlags As Int32, _
            ByRef StringLength As Int32, _
            ByVal LinkName As String _
        ) As SafeFindHandle

        Friend Declare Unicode Function FindFirstFileNameTransactedW Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal dwFlags As Int32, _
            ByRef StringLength As Int32, _
            ByVal LinkName As String, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As SafeFindHandle

        Friend Declare Unicode Function FindNextFileNameW Lib "Kernel32.dll" ( _
            ByVal hFindStream As SafeFindHandle, _
            ByRef StringLength As Int32, _
            ByVal LinkName As String _
        ) As Boolean
#End Region

#Region "Find*StreamW"
        Friend Declare Unicode Function FindFirstStreamW Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal InfoLevel As STREAM_INFO_LEVELS, _
            ByRef lpFindStreamData As WIN32_FIND_STREAM_DATA, _
            ByVal dwFlags As Int32 _
        ) As SafeFindHandle

        Friend Declare Unicode Function FindFirstStreamTransactedW Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal InfoLevel As STREAM_INFO_LEVELS, _
            ByRef lpFindStreamData As WIN32_FIND_STREAM_DATA, _
            ByVal dwFlags As Int32, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As SafeFindHandle

        Friend Declare Function FindNextStreamW Lib "Kernel32.dll" ( _
            ByVal hFindStream As SafeFindHandle, _
            ByRef lpFindStreamData As WIN32_FIND_STREAM_DATA _
        ) As Boolean

        <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Unicode)> _
        Friend Structure WIN32_FIND_STREAM_DATA
            Friend StreamSize As Int64
            <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=MAX_PATH + 36)> _
            Friend cStreamName As String
        End Structure

        Friend Enum STREAM_INFO_LEVELS
            FindStreamInfoStandard
        End Enum
#End Region

#Region "GetLongPathName"
        Friend Declare Auto Function GetLongPathName Lib "Kernel32.dll" ( _
            ByVal lpszShortPath As String, _
            ByVal lpszLongPath As String, _
            ByVal cchBuffer As Int32 _
        ) As Int32

        Friend Declare Auto Function GetLongPathNameTransacted Lib "Kernel32.dll" ( _
            ByVal lpszShortPath As String, _
            ByVal lpszLongPath As String, _
            ByVal cchBuffer As Int32, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Int32
#End Region

#Region "GetFullPathName"
        Friend Declare Auto Function GetFullPathName Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal nBufferLength As Int32, _
            ByVal lpBuffer As String, _
            ByRef lpFilePart As IntPtr _
        ) As Int32

        Friend Declare Auto Function GetFullPathNameTransacted Lib "Kernel32.dll" ( _
            ByVal lpFileName As String, _
            ByVal nBufferLength As Int32, _
            ByVal lpBuffer As String, _
            ByRef lpFilePart As IntPtr, _
            ByVal hTransaction As SafeTransactionHandle _
        ) As Int32
#End Region

#Region "Registry"
        Friend Declare Auto Function RegCreateKeyTransacted Lib "Advapi32.dll" ( _
            ByVal hKey As SafeKeyHandle, _
            ByVal lpSubKey As String, _
            ByVal Reserved As Int32, _
            ByVal lpClass As String, _
            ByVal dwOptions As REG_OPTION_Constants, _
            ByVal samDesired As REGSAM, _
            ByRef lpSecurityAttributes As SECURITY_ATTRIBUTES, _
            ByRef phkResult As SafeKeyHandle, _
            ByRef lpdwDisposition As REG_Disposition_Constants, _
            ByVal hTransaction As SafeTransactionHandle, _
            ByVal pExtendedParemeter As IntPtr _
        ) As Int32

        Friend Declare Auto Function RegOpenKeyTransacted Lib "Advapi32.dll" ( _
            ByVal hKey As SafeKeyHandle, _
            ByVal lpSubKey As String, _
            ByVal ulOptions As Int32, _
            ByVal samDesired As REGSAM, _
            ByRef phkResult As SafeKeyHandle, _
            ByVal hTransaction As SafeTransactionHandle, _
            ByVal pExtendedParemeter As IntPtr _
        ) As Int32

        Friend Declare Auto Function RegDeleteKeyTransacted Lib "Advapi32.dll" ( _
            ByVal hKey As SafeKeyHandle, _
            ByVal lpSubKey As String, _
            ByVal samDesired As REGSAM, _
            ByVal Reserved As Int32, _
            ByVal hTransaction As SafeTransactionHandle, _
            ByVal pExtendedParameter As IntPtr _
        ) As Int32

        Friend Shared ReadOnly HKEY_CLASSES_ROOT As New SafeKeyHandle(New IntPtr(&H80000000), False)
        Friend Shared ReadOnly HKEY_CURRENT_USER As New SafeKeyHandle(New IntPtr(&H80000001), False)
        Friend Shared ReadOnly HKEY_LOCAL_MACHINE As New SafeKeyHandle(New IntPtr(&H80000002), False)
        Friend Shared ReadOnly HKEY_USERS As New SafeKeyHandle(New IntPtr(&H80000003), False)
        Friend Shared ReadOnly HKEY_CURRENT_CONFIG As New SafeKeyHandle(New IntPtr(&H80000005), False)

        Friend Enum REG_OPTION_Constants
            REG_OPTION_BACKUP_RESTORE = &H4
            REG_OPTION_NON_VOLATILE = &H0
            REG_OPTION_VOLATILE = &H1
        End Enum

        Friend Enum REG_Disposition_Constants
            REG_CREATED_NEW_KEY = &H1
            REG_OPENED_EXISTING_KEY = &H2
        End Enum

        <Flags()> _
        Friend Enum REGSAM
            KEY_ALL_ACCESS = &HF003F
            KEY_CREATE_LINK = &H20
            KEY_CREATE_SUB_KEY = &H4
            KEY_ENUMERATE_SUB_KEYS = &H8
            KEY_EXECUTE = &H20019
            KEY_NOTIFY = &H10
            KEY_QUERY_VALUE = &H1
            KEY_READ = &H20019
            KEY_SET_VALUE = &H2
            KEY_WOW64_32KEY = &H200
            KEY_WOW64_64KEY = &H100
            KEY_WRITE = &H20006
        End Enum
#End Region

#Region "Miscellaneous functions"
        Friend Declare Function CloseHandle Lib "Kernel32.dll" ( _
            ByVal hObject As IntPtr _
        ) As Boolean
#End Region

#Region "Miscellaneous structures"
        <StructLayout(LayoutKind.Sequential)> _
        Friend Structure SECURITY_ATTRIBUTES
            Friend nLength As Int32
            Friend lpSecurityDescriptor As IntPtr
            Friend bInheritHandle As Boolean
        End Structure
#End Region

#Region "Miscellaneous constants"
        Friend Const MAX_PATH As Int32 = 260
        Friend Shared ReadOnly INVALID_HANDLE_VALUE As New IntPtr(-1)

        Friend Shared ReadOnly InvalidSafeFileHandle As New SafeFileHandle(INVALID_HANDLE_VALUE, False)
#End Region

#Region "Error codes"
        Friend Const NO_ERROR As Int32 = 0
        Friend Const ERROR_SUCCESS As Int32 = 0
        Friend Const ERROR_MORE_DATA As Int32 = 234
        Friend Const ERROR_HANDLE_EOF As Int32 = 38
#End Region

#Region "IKernelTransaction"
        < _
            ComImport(), _
            InterfaceType(ComInterfaceType.InterfaceIsIUnknown), _
            Guid("79427A2B-F895-40E0-BE79-B57DC82ED231") _
        > _
        Friend Interface IKernelTransaction
            '<PreserveSig()> _
            'Function GetHandle(<[Out]()> ByRef handle As IntPtr) As IntPtr
            Sub GetHandle(<[Out]()> ByRef handle As SafeTransactionHandle)
        End Interface

        Friend Shared ReadOnly S_OK As IntPtr = New IntPtr(1)
#End Region
    End Class
End Namespace