﻿Option Explicit On
Option Strict On

Imports System.Runtime.InteropServices
Imports System.Security.Permissions
Imports System.Transactions

Imports Microsoft.Win32.SafeHandles
Imports MvpsOrg.Dotnet.Win32.NativeMethods

Namespace MvpsOrg.Dotnet.Win32.SafeHandles
    <SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode:=True)> _
    Public NotInheritable Class SafeTransactionHandle
        Inherits SafeHandleZeroOrMinusOneIsInvalid

        ' Required by the p/invoke marshaller.
        Public Sub New()
            MyBase.New(True)
        End Sub

        Public Sub New( _
            ByVal preexistingHandle As IntPtr, _
            ByVal ownsHandle As Boolean _
        )
            MyBase.New(ownsHandle)
            Me.SetHandle(preexistingHandle)
        End Sub

        Protected Overrides Function ReleaseHandle() As Boolean
            Return CloseHandle(Me.handle)
        End Function

        Public Shared Function GetHandleFromDtcTransaction( _
            ByVal Transaction As Transaction _
        ) As SafeTransactionHandle
            Dim KernelTransaction As IKernelTransaction = _
                DirectCast( _
                    TransactionInterop.GetDtcTransaction( _
                        Transaction _
                    ),  _
                    IKernelTransaction _
                )
            Dim TransactionHandle As SafeTransactionHandle = Nothing

            ' If '<PreserveSig()>' was specified on 'GetHandle', we would have
            ' to throw the COM exception ourselves.
            '
            '    Dim Result As IntPtr = ...
            '    If Result <> S_OK Then
            '        Throw New COMException(Nothing, Result.ToInt32())
            '    End If
            KernelTransaction.GetHandle(TransactionHandle)
            Return TransactionHandle
        End Function
    End Class
End Namespace